/**
 * @file Controller logic for asr.xsl.
 * @author Carl Stephens
 */

import { TranscribeService, TranscriptionError, TranscriptionModel } from "./TranscribeModule.js";

/** @type {HTMLAudioElement} */
// @ts-ignore
const TRANSCRIPTION_AUDIO_ELEMENT = document.getElementById("transcriptionAudio");

/** @type {HTMLSourceElement} */
// @ts-ignore
const TRANSCRIPTION_AUDIO_SOURCE_ELEMENT = document.getElementById("transcriptionAudioSource");

/** @type {HTMLUListElement} */
// @ts-ignore
const TRANSCRIPTIONS_LIST = document.getElementById("transcriptionsList");

/** @type {HTMLTemplateElement} */
// @ts-ignore
const TRANSCRIPTION_TEMPLATE = document.getElementById("transcriptionTemplate");

let cachedAudioFileList = new Map();

// Get the size of each character in our monospace font
var MONOSPACE_CHAR_SIZE;
window.addEventListener("load", function()
{
    const monoCharSizeTestElement = document.querySelector(".monospace-font-size");
    if (monoCharSizeTestElement == null || monoCharSizeTestElement.textContent == null)
    {
        MONOSPACE_CHAR_SIZE = 8; // Slightly over-estimated guess for size 16 font.
        return;
    }
    
    MONOSPACE_CHAR_SIZE = monoCharSizeTestElement.clientWidth / monoCharSizeTestElement.textContent.length;
});

var transcribeService = new TranscribeService();

// @ts-ignore
const AudioUploadComponent = Vue.createApp(
    {
        data()
        {
            return {
                files: undefined,
                canTranscribe: false,
                isTranscribing: false
            }
        },
        methods:
        {
            openFilePicker() {
                document.getElementById("audioFileInput")?.click();
            },
            onFilesChanged()
            {
                /** @type {HTMLInputElement} */
                // @ts-ignore | Object could be null
                const audioFileInput = document.getElementById("audioFileInput");

                if (audioFileInput.files?.length != undefined && audioFileInput.files?.length > 0) {
                    this.canTranscribe = true;
                    this.files = audioFileInput?.files
                }
                else {
                    this.canTranscribe = false;
                    this.files = undefined;
                }
            },
            async doTranscription() {
                // TODO: Push files to queue. If currently transcribing, good to go; it'll pull off it.
                // Else call getTranscriptions();
                // Then, we don't have to worry about preventing the user from transcribing multiple items.
                await getTranscriptions();
            }
        }
    }
);
const AudioUploadVM = AudioUploadComponent.mount("#audioUploadContainer");

const TranscriptionsListComponent = Vue.createApp(
    {
        data()
        {
            return {
                /** @type {TranscriptionModel[]} */
                transcriptions: [],
                /** @type {TranscriptionError[]} */
                failures: []
            }
        },
        methods:
        {
            playAudioFile(fileName) {
                loadAudioFile(fileName);
                // TODO: play audio element
            },
            // TODO: give transcriptions a unique ID
            removeTranscription(id) {
                
            }
        }
    }
)
const TranscriptionsListVM = TranscriptionsListComponent.mount("#transcriptionsDisplayContainer");

/**
 * When awaited, creates a delay for the given number of milliseconds.
 * 
 * @param {Number} delayInms The number of milliseconds to delay for.
 * @returns A promise that will resolve when the delay has finished.
 */
function delay(delayInms)
{
    return new Promise(
        resolve =>
        {
            setTimeout(
                () => {
                    resolve(2);
                },
                delayInms
            );
        }
    );
}

/**
 * Gets the transcription of each submitted audio file.
 */
async function getTranscriptions()
{
    AudioUploadVM.isTranscribing = true;

    /** @type {FileList} */
    const files = AudioUploadVM.files;
    
    await delay(500); // TODO: Remove - UI testing purposes only
    // Cache the file list so that we can playback audio in the future
    for (const file of files) {
        cachedAudioFileList.set(file.name, file)
    }
    
    // Transcribe each audio file in batches.
    try
    {
        for await (const batch of transcribeService.batchTranscribeFiles(files))
        {
            for (const t of batch)
            {
                if (!t.success) {
                    TranscriptionsListVM.failures.push(new TranscriptionError(t.log, t.file_name));
                }
                else {
                    TranscriptionsListVM.transcriptions.push(t);
                }
            }
        }
    }
    catch (e)
    {
        console.error("Failed to transcribe files");
        console.error(e);
        TranscriptionsListVM.failures.push(e);
    }
    
    AudioUploadVM.isTranscribing = false;
}

/**
* Inserts a transcription object into the DOM.
* Adapted from https://developer.mozilla.org/en-US/docs/Web/HTML/Element/template
* 
* @param {TranscriptionModel} transcription The transcription to insert.
*/
function insertTranscription(transcription)
{
    // Create a new transcription row
    /** @type {HTMLLIElement} */
    // @ts-ignore
    const clone = TRANSCRIPTION_TEMPLATE.content.firstElementChild.cloneNode(true);

    // @ts-ignore
    clone.querySelector(".transcription__text").textContent = transcription.transcription;
    // @ts-ignore
    clone.querySelector(".transcription__file-name").textContent = transcription.file_name;
    
    // Hook up the remove button
    /** @type {HTMLButtonElement} */
    // @ts-ignore
    const removeButton = clone.querySelector(".transcription__remove-button");
    removeButton.addEventListener("click", onDeleteTranscription);
    
    loadAudioFile(transcription.file_name);
    
    // Prepare the audio slider
    /** @type {HTMLInputElement} */
    // @ts-ignore
    const audioSlider = clone.querySelector(".audio-slider");
    audioSlider.max = TRANSCRIPTION_AUDIO_ELEMENT.duration.toString();
    audioSlider.step = "0.01";
    
    // Set the filename data property on every element in the clone's DOM node
    recurseAddData(clone, "file-name", transcription.file_name);

    // Insert the entry. This could occur first (if you need to make calculations based off of the rendered element)
    // But it is preferable to do it last in order to avoid multiple render passes.
    TRANSCRIPTIONS_LIST.appendChild(clone);

    const TranscriptionWordListComponent = Vue.createApp(
    {
        data()
        {
            return {
                lines: []
            }
        }
    });
    const TranscriptionWordListVM = TranscriptionWordListComponent.mount("#transcriptionWordList");

    transcription.transcription.replace('', "\u00A0"); // This helps with formatting, as whitespace is trimmed.

    // Get the amount of space that we have to align words within
    /** @type {HTMLDivElement} */
    // @ts-ignore
    const transcriptionTextList = clone.querySelector(".transcription__word-list");
    const charsPerLine = Math.floor(transcriptionTextList.clientWidth / MONOSPACE_CHAR_SIZE);
    
    // Insert words with correct line alignment
    const tText = transcription.transcription;

    for (let i = 0; i < tText.length; i += 0)
    {
        /** @type {String} */
        let slice;

        if (i + charsPerLine < tText.length)
        {
            slice = tText.slice(i, charsPerLine);

            if (tText.charAt(i) != ' ' && tText.charAt(i + 1) != ' ') {
                let lastSpacePos = slice.lastIndexOf(' ');
                let decrement = slice.length - lastSpacePos;
                slice = slice.slice(0, lastSpacePos);
                i -= decrement;
            }
        }
        else
        {
            slice = tText.slice(i);
        }

        TranscriptionWordListVM.lines.push({ words: slice.split(' ') });
        i += charsPerLine;
    }
}

/**
 * Loads an audio file.
 * @param {String} requestedAudioFile The name of the requested audio file.
 */
function loadAudioFile(requestedAudioFile)
{
    const currentAudioFile = TRANSCRIPTION_AUDIO_SOURCE_ELEMENT.dataset.fileName;

    // Load the appropiate audio if necessary.
    if (currentAudioFile != requestedAudioFile)
    {
        // If an audio file is already loaded we can revoke it.
        if (currentAudioFile) {
            URL.revokeObjectURL(currentAudioFile);
        }

        const urlObject = URL.createObjectURL(cachedAudioFileList.get(requestedAudioFile));
        TRANSCRIPTION_AUDIO_SOURCE_ELEMENT.src = urlObject;
        TRANSCRIPTION_AUDIO_SOURCE_ELEMENT.dataset.fileName = requestedAudioFile;
        TRANSCRIPTION_AUDIO_ELEMENT.load();
    }
}

/**
 * Removes a transcription from the DOM.
 * 
 * @param {MouseEvent} ev The mouse click event.
 */
function onDeleteTranscription(ev)
{
    if (ev == null || ev.target == null) {
        return;
    }

    const fileName = ev.target.dataset.fileName;
    const child = document.querySelector(`.transcription__container[data-file-name='${fileName}']`);

    TRANSCRIPTIONS_LIST.removeChild(child);
    cachedAudioFileList.delete(fileName);
}

/**
 * Recurses through the entire tree of a parent and adds the given data attribute.
 * 
 * @param {Element} parent The parent node.
 * @param {String} dataElementName The name of the data element. Must use hyphen notation, rather than camel case.
 * @param {String} value The value of the data element.
 */
function recurseAddData(parent, dataElementName, value)
{
    parent.setAttribute("data-" + dataElementName, value);

    for (const child of parent.children) {
        recurseAddData(child, dataElementName, value);
    }
}